/*
 * Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
 * Copyright (c) 1999 Ben Williamson <benw@pobox.com>
 */

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <termios.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>
#include <errno.h>
#include "serial.h"
#include "util.h"
#include "log.h"

#define SERIAL_BLOCKSIZE	0x1000

static struct termios oldtio, newtio, contio;
static int portfd = -1;

//prototype declaration
static int xread(int fd, void *_buf, int count);
//static int xaread(int fd, void *_buf, int count);
static int xwrite(int fd, const void *_buf, int count);
static int xawrite(int fd, const void *_buf, int count);

/* kill handler used when only the serial port is open */
static void handler1(int signal){
  if(portfd >= 0){
	tcsetattr(portfd, TCSANOW, &oldtio);
	serial_close();
  }
  exit(1);
}

/* kill handler used in terminal mode */
static void handler2(int signal){
  if(portfd >= 0){
	tcsetattr(portfd, TCSANOW, &oldtio);
	tcsetattr(STDIN_FILENO, TCSANOW, &contio);
	serial_close();
  }
  exit(1);
}

/* open the serial port */
int serial_open(const char *dev){
  int ret;
	portfd = open(dev, O_RDWR | O_NOCTTY);
	if(portfd < 0){
		print_perror(dev);
		return -1;
	}

	/* save current port settings */
	ret = tcgetattr(portfd, &oldtio);
	if(ret < 0){
		print_perror("tcgetattr");
		return -1;
	}

	/* set signal handlers before switching settings */
	signal(SIGHUP, handler1);
	signal(SIGINT, handler1);
	signal(SIGPIPE, handler1);
	signal(SIGTERM, handler1);

	/* configure new port settings: 9600 8N1 */
	memset(&newtio, 0, sizeof(newtio));
	newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR;
	newtio.c_oflag = 0;
	/* set input mode (non-canonical, no echo,...) */
	newtio.c_lflag = 0;
	newtio.c_cc[VTIME] = 0;	/* inter-character timer unused */
	newtio.c_cc[VMIN] = 1;	/* blocking read until 1 char received */

	/* install new port settings */
	tcflush(portfd, TCIFLUSH);
	ret = tcsetattr(portfd, TCSANOW, &newtio);
	if(ret < 0){
		print_perror("tcsetattr");
		return -1;
	}
	return 0;
}

/* close serial port and restore settings */
int serial_close(void){
  int ret;
  if(portfd >= 0){
	tcsetattr(portfd, TCSANOW, &oldtio);
	ret = close(portfd);
	portfd = -1;
	if(ret < 0){
		print_perror("close");
		return -1;
	}
  }
	return 0;
}

/* switch baud rate */

int serial_baud(int baud){
  speed_t speed;

  switch(baud){
  case 115200: speed = B115200; break;
  default:     speed = B115200; break;
  }

  if(portfd >= 0){
	tcdrain(portfd);
	usleep(50 * 1000);	/* 50 ms sleep; arbitrary */
	cfsetispeed(&newtio, speed);
	cfsetospeed(&newtio, speed);
	tcsetattr(portfd, TCSANOW, &newtio);
	return 0;
  }
  return -1;
}

/* enter terminal mode */
void serial_terminal(void){
	struct termios tio;
		
	message("Pretending to be a terminal - interrupt to exit.\n");
	tcgetattr(STDIN_FILENO, &contio);
	signal(SIGHUP, handler2);
	signal(SIGINT, handler2);
	signal(SIGPIPE, handler2);
	signal(SIGTERM, handler2);
	/* set input mode (non-canonical, no echo,...) */
	memcpy(&tio, &contio, sizeof tio);
	tio.c_lflag = ISIG;
	tio.c_cc[VTIME] = 0;   /* inter-character timer unused */
	tio.c_cc[VMIN] = 1;   /* blocking read until 1 char received */
	tcsetattr(STDIN_FILENO,TCSANOW,&tio);

	while(1){
		fd_set fds;
		int retval;
		unsigned char c;
			
		FD_ZERO(&fds);
		FD_SET(STDIN_FILENO, &fds);
		FD_SET(portfd, &fds);

		retval = select(portfd + 1, &fds, NULL, NULL, NULL);
		if (FD_ISSET(portfd, &fds)) {
			if (1 == read(portfd, &c, 1))
				putchar(c);
		}
		if (FD_ISSET(STDIN_FILENO, &fds)) {
			if (1 == read(STDIN_FILENO, &c, 1))
				put_char(c);
		}
		fflush(NULL);
	}
}

/* wait for a character on the serial port */
unsigned char get_char(void){
	unsigned char c;
	if(portfd < 0){
	  message("Not open Serial-Port\n");
	  return -1;
	}
	xread(portfd, &c, 1);
	return c;
}

/* wait for a character, or until a given timeout */
int get_char_timeout(int msecs){
	fd_set fds;
	struct timeval tv;

	if(portfd < 0){
	  message("Not open Serial-Port\n");
	  return -1;
	}

	FD_ZERO(&fds);
	FD_SET(portfd, &fds);
	tv.tv_sec = 0;
	tv.tv_usec = msecs * 1000;

	select(portfd+1, &fds, NULL, NULL, &tv);
	if (FD_ISSET(portfd, &fds))
		return get_char();
	return -1;
}

/* send a character on the serial port */
int put_char(unsigned char c){
  if(portfd < 0){
	  message("Not open Serial-Port\n");
	  return -1;
  }
  xwrite(portfd, &c, 1);
  return 0;
}

/* wait for a word on the serial port, LSB first */
unsigned long get_word(void){
	unsigned long w;
	
	w = get_char() & 0xff;
	w |= (get_char() & 0xff) << 8;
	w |= (get_char() & 0xff) << 16;
	w |= (get_char() & 0xff) << 24;
	return w;
}

/* send a word on the serial port, little-endian */
int put_word(unsigned long w){
	put_char((unsigned char)(w));
	put_char((unsigned char)(w >> 8));
	put_char((unsigned char)(w >> 16));
	put_char((unsigned char)(w >> 24));
	return 0;
}

int put_block(const char *buf, int size){
  if(portfd < 0){
	  message("Not open Serial-Port\n");
	  return -1;
  }
  xawrite(portfd, buf, size);
  return 0;
}

/* ask the target to read a byte of memory */
unsigned char target_read_byte(unsigned long addr){
	put_char('g');
	put_word(addr);
	return get_char();
}

/* tell the target to write a byte of memory */
int target_write_byte(unsigned long addr, unsigned char data){
	put_char('s');
	put_word(addr);
	put_char(data);
	return 0;
}

/* ask the target to read a word of memory */
unsigned long target_read_word(unsigned long addr){
	put_char('r');
	put_word(addr);
	return get_word();
}

/* tell the target to write a word of memory */
int target_write_word(unsigned long addr, unsigned long data){
	put_char('w');
	put_word(addr);
	put_word(data);
	return 0;
}

/* tell the target to write a block of memory */
int target_write_block(unsigned long addr, const char *buf,
			int size, int progress)
{
	char str[256];
	if(portfd < 0){
	  	  message("Not open Serial-Port\n");
	  return -1;
	}

	while (size > 0) {
		char checksum = 0;
		int step = min(size, SERIAL_BLOCKSIZE);
		
		put_char('W');
		put_word(addr);
		put_word(step);
		xawrite(portfd, buf, step);
		addr += step;
		size -= step;
		progress += step;
		while (step-- > 0) {
			checksum += *buf++;
		}
		if (checksum != (char)get_char()) {
			message("\nSerial checksum error\n");
			messagebox("\nSerial checksum error\n");
			return -1;
		}
		
		sprintf(str, "0x%08x\r", progress);
		message(str);
		fflush(NULL);
	}
	return 0;
}

static int xread(int fd, void *_buf, int count){
  char *buf = (char *)_buf;
  int ret;

  for(;;){
    ret = read(fd, buf, count);
    if(ret < 0 && errno == EINTR){
      continue;
    }
    if(ret < 0){
      print_perror("read");
      return -1;
    }
    break;
  }
  return ret;
}

#if 0
static int xaread(int fd, void *_buf, int count){
  char *buf = (char *)_buf;
  int readlen;
  int ret;

  for(readlen = 0;;){
    ret = read(fd, &buf[readlen], count - readlen);
    if(ret < 0 && errno == EINTR){
      continue;
    }
    if(ret < 0){
      print_perror("read");
      return -1;
    }
    readlen += ret;
    if(readlen == count) break;
  }
  return count;
}
#endif

static int xwrite(int fd, const void *_buf, int count){
  char *buf = (char *)_buf;
  int ret;

  for(;;){
    ret = write(fd, buf, count);
    if(ret < 0 && errno == EINTR){
      continue;
    }
    if(ret < 0){
      print_perror("write");
      return -1;
    }
    break;
  }
  return ret;
}

static int xawrite(int fd, const void *_buf, int count){
	char *buf = (char *)_buf;
	int writelen;
	int ret;

	for(writelen = 0;;){
	  ret = write(fd, &buf[writelen], count - writelen);
	  if(ret < 0 && errno == EINTR){
	    continue;
	  }
	  if(ret < 0){
	    print_perror("write");
	    return -1;
	  }
	  writelen += ret;
	  if(writelen == count) break;
	}
	return count;
}

